from .resource import Resource
from .resources.organization import Organization
from .resources.cron import Cron
from .resources.user import User
from .resources.plan import Plan
from .resources.chomp_task import ChompTask
from .resources.mobius_task import MobiusTask


class Client(Resource):
    def get_organizations(self, limit=25, offset=0, **kwargs):
        return Organization.get_all(parent=self,
                                    limit=limit,
                                    offset=offset,
                                    **kwargs)

    def get_organization(self, id):
        return Organization.get(parent=self, id=id)

    def create_organization(self, **kwargs):
        return Organization.create(parent=self, **kwargs)

    def cron(self):
        """Internal only - cron job manual timer"""
        return Cron.get_all(parent=self)

    def get_users(self, limit=25, offset=0, **kwargs):

        # Some supported filters: filterbyUsername, filterByEmail
        return User.get_all(parent=self, limit=limit, offset=offset, **kwargs)

    def get_user(self, id):
        return User.get(parent=self, id=id)

    def get_plans(self, **kwargs):
        return Plan.get_all(parent=self, **kwargs)

    def get_plan(self, id):
        return Plan.get(parent=self, id=id)

    def get_chomp_task(self, id):
        # id is schedule id
        return ChompTask.get(parent=self)

    def get_chomp_tasks(self, **kwargs):
        return ChompTask.get(parent=self)

    def claim_chomp_task(self):
        return ChompTask.create(parent=self)

    def get_mobius_task(self, id):
        # id is schedule id
        return MobiusTask.get(parent=self)

    def get_mobius_tasks(self, **kwargs):
        return MobiusTask.get(parent=self)

    def claim_mobius_task(self):
        return MobiusTask.create(parent=self)



from ..resource import Resource
from .worker import Worker
from .schedule import Schedule
from .shift import Shift
from .recurring_shift import RecurringShift


class Role(Resource):
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}"
    ID_NAME = "role_id"

    def get_workers(self, **kwargs):
        return Worker.get_all(parent=self, **kwargs)

    def get_worker(self, id=id):
        return Worker.get(parent=self, id=id)

    def create_worker(self, **kwargs):
        return Worker.create(parent=self, **kwargs)

    def get_schedules(self, **kwargs):
        return Schedule.get_all(parent=self)

    def get_schedule(self, id):
        return Schedule.get(parent=self, id=id)

    def get_shifts(self, **kwargs):
        return Shift.get_all(parent=self, **kwargs)

    def get_shift(self, id):
        return Shift.get(parent=self, id=id)

    def create_shift(self, **kwargs):
        return Shift.create(parent=self, **kwargs)

    def get_recurring_shifts(self, **kwargs):
        return RecurringShift.get_all(parent=self, **kwargs)

    def get_recurring_shift(self, id):
        return RecurringShift.get(parent=self, id=id)

    def create_recurring_shift(self, **kwargs):
        return RecurringShift.create(parent=self, **kwargs)



from ..resource import Resource


class ShiftEligibleWorker(Resource):
    """Returns workers that can claim a given shift"""
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/shifts/{shift_id}/users/{user_id}"
    ID_NAME = "user_id"



from ..resource import Resource


class Timeclock(Resource):
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/users/{user_id}/timeclocks/{timeclock_id}"
    ID_NAME = "timeclock_id"



import os

import pytest

from staffjoy import Client, UnauthorizedException

from . import logger
"""
This test file is intended for use in continuous integration. It runs
against the Staging environment of Staffjoy in a dedicated functional
testing organization. We will not be giving public access to Staffjoy
stage, but you can modify this script to run against your own org.
For a developer access, please email help@staffjoy.com
"""

TEST_ORG = 18
ENV = "stage"
KEY = os.environ.get("STAFFJOY_STAGE_API_KEY")
TEST_WORKER = "feynman@7bridg.es"


def test_org_crud():
    c = Client(key=KEY, env=ENV)

    # Just some basic stuff
    assert len(c.get_plans()) > 0

    logger.debug("Fetching organization")
    o = c.get_organization(TEST_ORG)

    assert o.get_id() == TEST_ORG

    location_count = len(o.get_locations())

    logger.debug("Changing organization name")
    o.patch(name="[In Progress] Continuous integration test")

    logger.debug("Creating a location")
    l = o.create_location(name="El Farolito", timezone="America/Los_Angeles")
    l_id = l.get_id()
    logger.debug("Location id {}".format(l_id))

    assert l.data.get("name") == "El Farolito"
    logger.debug("Changing location name")
    l.patch(name="La Taqueria")

    logger.debug("Checking that location is created")
    new_location_count = len(o.get_locations())
    assert new_location_count == (location_count + 1)
    del l

    logger.debug("Fetching location by ID")
    l = o.get_location(l_id)
    assert l.data.get("name") == "La Taqueria"

    logger.debug("Testing role crud")
    r = l.create_role(name="Kitchen")
    r.patch(name="Cocina")
    logger.debug("Adding worker")
    r.get_workers()
    r.create_worker(email=TEST_WORKER,
                    min_hours_per_workweek=30,
                    max_hours_per_workweek=40)

    logger.debug("Deleting worker")
    r.delete()

    logger.debug("Deleting location")
    l.delete()
    del l
    logger.debug("Making sure location has been archived")

    loc = o.get_location(l_id)
    assert loc.data.get("archived")

    logger.debug("Finishing up")
    o.patch(name="Continuous integration test")
    all_locations = o.get_locations()
    for location in all_locations:
        if not location.data.get("archived"):
            location.delete()

    for location in o.get_locations():
        assert location.data.get("archived")



import logging


class DefaultConfig:
    ENV = "prod"
    LOG_LEVEL = logging.INFO
    BASE = "https://www.staffjoy.com/api/v2/"
    REQUEST_SLEEP = 0.5


class StageConfig(DefaultConfig):
    ENV = "stage"
    BASE = "https://stage.staffjoy.com/api/v2/"


class DevelopmentConfig(DefaultConfig):
    ENV = "dev"
    LOG_LEVEL = logging.DEBUG
    BASE = "http://dev.staffjoy.com/api/v2/"


config_from_env = {  # Determined in main.py
    "dev": DevelopmentConfig,
    "stage": StageConfig,
    "prod": DefaultConfig,
}



from ..resource import Resource
from .shift_eligible_workers import ShiftEligibleWorker


class Shift(Resource):
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/shifts/{shift_id}"
    ID_NAME = "shift_id"

    def get_eligible_workers(self, **kwargs):
        return ShiftEligibleWorker.get_all(self, **kwargs)



from ..resource import Resource


class Preference(Resource):
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/schedules/{schedule_id}/preferences/{user_id}"
    ID_NAME = "user_id"



from ..resource import Resource
from .role import Role


class Location(Resource):
    PATH = "organizations/{organization_id}/locations/{location_id}"
    ID_NAME = "location_id"

    def get_roles(self):
        return Role.get_all(parent=self)

    def get_role(self, id):
        return Role.get(parent=self, id=id)

    def create_role(self, **kwargs):
        return Role.create(parent=self, **kwargs)



from ..resource import Resource
from .session import Session
from .apikey import ApiKey


class User(Resource):
    PATH = "users/{user_id}"
    ID_NAME = "user_id"

    def get_sessions(self):
        return Session.get_all(parent=self)

    def get_session(self, id):
        return Session.get(parset=self, id=id)

    def get_apikeys(self):
        return ApiKey.get_all(parent=self)

    def get_apikey(self, id):
        return ApiKey.get(parent=self, id=id)



from ..resource import Resource


class TimeOffRequest(Resource):
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/users/{user_id}/timeoffrequests/{time_off_request_id}"
    ID_NAME = "time_off_request_id"



from ..resource import Resource
from .location import Location
from .admin import Admin
from .organization_worker import OrganizationWorker


class Organization(Resource):
    PATH = "organizations/{organization_id}"
    ID_NAME = "organization_id"

    def get_locations(self):
        return Location.get_all(parent=self)

    def get_location(self, id):
        return Location.get(parent=self, id=id)

    def create_location(self, **kwargs):
        return Location.create(parent=self, **kwargs)

    def get_admins(self):
        return Admin.get_all(parent=self)

    def get_admin(self, id):
        return Admin.get(parent=self, id=id)

    def create_admin(self, **kwargs):
        """Typically just pass email"""
        return Admin.create(parent=self, **kwargs)

    def get_workers(self, **kwargs):
        return OrganizationWorker.get_all(parent=self, **kwargs)



from ..resource import Resource


class ScheduleTimeOffRequest(Resource):
    """this is only a get collection endpoint"""
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/schedules/{schedule_id}/timeoffrequests/"



from ..resource import Resource


class RecurringShift(Resource):
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/recurringshifts/{recurring_shift_id}"
    ID_NAME = "recurring_shift_id"



from ..resource import Resource


class ScheduleTimeclock(Resource):
    """this is only a get collection endpoint"""
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/schedules/{schedule_id}/timeclocks/"



class UnauthorizedException(Exception):
    def __init__(self, code=401, response={}):

        self.code = code
        # We envelope error messages like this
        self.message = response.get("message", "Unknown Error")
        super(UnauthorizedException, self).__init__()

    def __str__(self):
        return "\n Code: {} \n Message: {} \n".format(self.code, self.message)


class NotFoundException(Exception):
    def __init__(self, code=404, response={}):

        self.code = code
        # We envelope error messages like this
        self.message = response.get("message", "Unknown Error")
        super(NotFoundException, self).__init__()

    def __str__(self):
        return "\n Code: {} \n Message: {} \n".format(self.code, self.message)


class BadRequestException(Exception):
    def __init__(self, code=400, response={}):

        self.code = code
        # We envelope error messages like this
        self.message = response.get("message", "Unknown Error")
        super(BadRequestException, self).__init__()

    def __str__(self):
        return "\n Code: {} \n Message: {} \n".format(self.code, self.message)



import requests
import time
from copy import copy

from .config import config_from_env
from .exceptions import UnauthorizedException, NotFoundException, BadRequestException


class Resource:
    # Seconds to sleep between requests (bc of rate limits)
    REQUEST_SLEEP = 0.5

    PATH = ""  # URL path added to base, including route variables
    ID_NAME = None  # What is this ID called in the route of children?
    META_ENVELOPES = []  # Metadata keys for what to unpack from response
    ENVELOPE = "data"  # We "envelope" response data in the "data" section
    TRUTHY_CODES = [requests.codes.ok, requests.codes.created,
                    requests.codes.no_content, requests.codes.accepted]

    def __init__(self,
                 key="",
                 config=None,
                 env="prod",
                 data={},
                 route={},
                 meta={}):
        """Initialize the resource"""
        self.key = key

        self.config = config or config_from_env.get(env, "prod")

        # These should be overridden by child classes
        self.data = data  # Data from the read method
        self.route = route  # Route variables
        self.meta = meta  # Meta data

    @classmethod
    def get(cls, parent=None, id=None, data=None):
        """Inherit info from parent and return new object"""
        # TODO - allow fetching of parent based on child?

        if parent is not None:
            route = copy(parent.route)
        else:
            route = {}

        if id is not None and cls.ID_NAME is not None:
            route[cls.ID_NAME] = id

        obj = cls(key=parent.key, route=route, config=parent.config)

        if data:
            # This is used in "get all" queries
            obj.data = data
        else:
            obj.fetch()

        return obj

    @classmethod
    def get_all(cls, parent=None, **params):

        if parent is not None:
            route = copy(parent.route)
        else:
            route = {}
        if cls.ID_NAME is not None:
            # Empty string triggers "get all resources"
            route[cls.ID_NAME] = ""

        base_obj = cls(key=parent.key, route=route, config=parent.config)
        """Perform a read request against the resource"""

        r = requests.get(base_obj._url(),
                         auth=(base_obj.key, ""),
                         params=params)
        time.sleep(cls.REQUEST_SLEEP)

        if r.status_code not in cls.TRUTHY_CODES:
            return base_obj._handle_request_exception(r)

        response = r.json()
        objects_data = response.get(base_obj.ENVELOPE or base_obj, [])

        return_objects = []
        for data in objects_data:
            # Note that this approach does not get meta data
            return_objects.append(cls.get(parent=parent,
                                          id=data.get(cls.ID_NAME, data.get(
                                              "id")),
                                          data=data))

        return return_objects

    def _url(self):
        """Get the URL for the resource"""
        if self.ID_NAME not in self.route.keys() and "id" in self.data.keys():
            self.route[self.ID_NAME] = self.data["id"]
        return self.config.BASE + self.PATH.format(**self.route)

    @staticmethod
    def _handle_request_exception(request):
        """Raise the proper exception based on the response"""
        try:
            data = request.json()
        except:
            data = {}

        code = request.status_code
        if code == requests.codes.bad:
            raise BadRequestException(response=data)

        if code == requests.codes.unauthorized:
            raise UnauthorizedException(response=data)

        if code == requests.codes.not_found:
            raise NotFoundException(response=data)

        # Generic error fallback
        request.raise_for_status()

    def fetch(self):
        """Perform a read request against the resource"""
        r = requests.get(self._url(), auth=(self.key, ""))
        time.sleep(self.REQUEST_SLEEP)

        if r.status_code not in self.TRUTHY_CODES:
            return self._handle_request_exception(r)

        response = r.json()
        if self.ENVELOPE:
            self.data = response.get(self.ENVELOPE, {})
        else:
            self.data = response

        # Move to separate function so it can be overrridden
        self._process_meta(response)

    def _process_meta(self, response):
        """Process additional data sent in response"""
        for key in self.META_ENVELOPES:
            self.meta[key] = response.get(key)

    def delete(self):
        """Delete the object"""

        r = requests.delete(self._url(), auth=(self.key, ""))
        time.sleep(self.REQUEST_SLEEP)

        if r.status_code not in self.TRUTHY_CODES:
            return self._handle_request_exception(r)

    def patch(self, **kwargs):
        """Change attributes of the item"""
        r = requests.patch(self._url(), auth=(self.key, ""), data=kwargs)
        time.sleep(self.REQUEST_SLEEP)

        if r.status_code not in self.TRUTHY_CODES:
            return self._handle_request_exception(r)

        # Refetch for safety. We could modify based on response,
        # but I'm afraid of some edge cases and marshal functions.
        self.fetch()

    @classmethod
    def create(cls, parent=None, **kwargs):
        """Create an object and return it"""

        if parent is None:
            raise Exception("Parent class is required")

        route = copy(parent.route)
        if cls.ID_NAME is not None:
            route[cls.ID_NAME] = ""

        obj = cls(key=parent.key, route=route, config=parent.config)

        response = requests.post(obj._url(), auth=(obj.key, ""), data=kwargs)
        time.sleep(cls.REQUEST_SLEEP)

        if response.status_code not in cls.TRUTHY_CODES:
            return cls._handle_request_exception(response)

        # No envelope on post requests
        data = response.json()
        obj.route[obj.ID_NAME] = data.get("id", data.get(obj.ID_NAME))
        obj.data = data

        return obj

    def get_id(self):
        return self.data.get("id", self.route.get(self.ID_NAME))

    def __str__(self):
        return "{} id {}".format(self.__class__.__name__,
                                 self.route.get(self.ID_NAME))



from setuptools import setup, find_packages

version = "0.16"
setup(name="staffjoy",
      packages=find_packages(),
      version=version,
      description="Staffjoy API Wrapper in Python",
      author="Philip Thomas",
      author_email="help@staffjoy.com",
      license="MIT",
      url="https://github.com/staffjoy/client_python",
      download_url="https://github.com/StaffJoy/client_python/archive/%s.tar.gz" % version,
      keywords=["staffjoy-api", "staffjoy", "staff joy"],
      install_requires=["requests[security]"], )



from ..resource import Resource
from .preference import Preference
from .schedule_shift import ScheduleShift
from .schedule_timeclock import ScheduleTimeclock
from .schedule_time_off_request import ScheduleTimeOffRequest


class Schedule(Resource):
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/schedules/{schedule_id}"
    ID_NAME = "schedule_id"

    def get_preferences(self, **kwargs):
        return Preference.get_all(parent=self, **kwargs)

    def get_preference(self, id):
        """ Get a worker's preference for a given week"""
        return Preference.get(parent=self, id=id)

    def create_preference(self, **kwargs):
        return Preference.create(parent=self, **kwargs)

    def get_schedule_shifts(self, **kwargs):
        return ScheduleShift.get_all(parent=self, **kwargs)

    def get_schedule_timeclocks(self, **kwargs):
        return ScheduleTimeclock.get_all(parent=self, **kwargs)

    def get_schedule_time_off_requests(self, **kwargs):
        return ScheduleTimeOffRequest.get_all(parent=self, **kwargs)



from ..resource import Resource
from .timeclock import Timeclock
from .time_off_request import TimeOffRequest


class Worker(Resource):
    """Organization administrators"""
    PATH = "organizations/{organization_id}/locations/{location_id}/roles/{role_id}/users/{user_id}"
    ID_NAME = "user_id"

    def get_timeclocks(self, **kwargs):
        return Timeclock.get_all(parent=self, **kwargs)

    def get_timeclock(self, id):
        return Timeclock.get(parent=self, id=id)

    def create_timeclock(self, **kwargs):
        return Timeclock.create(parent=self, **kwargs)

    def get_time_off_requests(self, **kwargs):
        return TimeOffRequest.get_all(parent=self, **kwargs)

    def get_time_off_request(self, id):
        return TimeOffRequest.get(parent=self, id=id)

    def create_time_off_request(self, **kwargs):
        return TimeOffRequest.create(parent=self, **kwargs)
